#include <dbi/dbi.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>

#include "proc_lua.h"
#include "log.h"

void
dump_stack(lua_State *L)
{
	int i;
	for (i = 1; i <= lua_gettop(L); i++) {
		printf("%d: (%s) ", i, lua_typename(L, lua_type(L, i)));
		switch (lua_type(L, i)) {
			case LUA_TNUMBER:
				printf("%g\n", lua_tonumber(L, i));
				break;
			case LUA_TBOOLEAN:
				printf("%s\n", lua_toboolean(L, i) ? "true" : "false");
				break;
			case LUA_TSTRING:
				printf("%s\n", lua_tostring(L, i));
				break;
			case LUA_TTABLE:
				lua_pushnil(L);
				while (lua_next(L, i) != 0) {
					switch(lua_type(L, -2)) {
						case LUA_TNUMBER:
							printf("%g", lua_tonumber(L, -2));
							break;
						case LUA_TSTRING:
							printf("%s", lua_tostring(L, -2));
							break;
						default:
							printf("(%s)",  lua_typename(L, lua_type(L, -2)));
							break;
					}
					printf(" = ");
					switch(lua_type(L, -1)) {
						case LUA_TNUMBER:
							printf("%g", lua_tonumber(L, -1));
							break;
						case LUA_TSTRING:
							printf("%s", lua_tostring(L, -1));
							break;
						default:
							printf("(%s)",  lua_typename(L, lua_type(L, -1)));
							break;
					}
					printf(", ");
					lua_pop(L, 1);
				}
				printf("\n");
				break;
			case LUA_TFUNCTION:
			case LUA_TUSERDATA:
			case LUA_TTHREAD:
			case LUA_TLIGHTUSERDATA:
			case LUA_TNIL:
				printf("\n");
		}
	}
}

static int
trap_call_cb(lua_State *L)
{
	// | 1=table | 2=key
	lua_getfield(L, 2, "trim");
	// | 1=table | 2=key | 3=trim()
	lua_insert(L, 2);
	// | 1=table | 2=trim() | 3=key
	lua_call(L, 1, 1);
	// | 1=table | 2=trimed key
	lua_getfield(L, 1, "fields");
	// | 1=table | 2=key | 3=fields
	lua_insert(L, 2);
	// | 1=table | 2=fields | 3=key
	lua_gettable(L, 2);
	// | 1=table | 2=fields | 3=value
	return 1;
}

static int
invalid_field_cb(lua_State *L)
{
	const char *field = lua_tostring(L, 1);
	return luaL_error(L, "trap does not contain field %s", field);
}

static int
trap_as_table(lua_State *L, proc_trap *trap)
{
	proc_field *f;

	lua_newtable(L);
	int trap_index = lua_gettop(L);

	lua_pushnumber(L, trap->timestamp);
	lua_setfield(L, trap_index, "timestamp");

	struct tm *tm = localtime(&trap->timestamp);
	lua_newtable(L);
	int time = lua_gettop(L);
	lua_pushnumber(L, tm->tm_hour);
	lua_setfield(L, time, "hour");
	lua_pushnumber(L, tm->tm_min);
	lua_setfield(L, time, "minute");
	lua_pushnumber(L, tm->tm_sec);
	lua_setfield(L, time, "second");
	lua_setfield(L, trap_index, "time");

	lua_newtable(L);
	int date = lua_gettop(L);
	lua_pushnumber(L, tm->tm_year);
	lua_setfield(L, date, "year");
	lua_pushnumber(L, tm->tm_mon);
	lua_setfield(L, date, "month");
	lua_pushnumber(L, tm->tm_mday);
	lua_setfield(L, date, "day");
	lua_pushnumber(L, tm->tm_wday);
	lua_setfield(L, date, "weekday");
	lua_setfield(L, trap_index, "date");

	lua_pushnumber(L, trap->id);
	lua_setfield(L, trap_index, "id");
	lua_pushstring(L, trap->host);
	lua_setfield(L, trap_index, "host");
	lua_pushstring(L, trap->oid);
	lua_setfield(L, trap_index, "oid");

	lua_getglobal(L, "constant");
	lua_newtable(L);
	int fields = lua_gettop(L);
	for (f = trap->fields; f; f = f->next) {
		switch(f->type) {
			case FIELD_STRING:
				lua_pushstring(L, f->value.string);
				break;
			case FIELD_NUMERIC:
				lua_pushnumber(L, f->value.number);
				break;
			case FIELD_BOOLEAN:
				lua_pushboolean(L, f->value.boolean);
				break;
			case FIELD_NULL:
				lua_pushnil(L);
				break;
			case FIELD_TIME:
				lua_pushnumber(L, f->value.number);
				break;
		}
		lua_setfield(L, fields, f->oid);
	}
	lua_newtable(L);
	int metatable = lua_gettop(L);
	lua_pushcfunction(L, invalid_field_cb);
	lua_setfield(L, metatable, "__index");
	lua_setmetatable(L, fields);
	lua_call(L, 1, 1);
	lua_setfield(L, trap_index, "fields");

	lua_newtable(L);
	metatable = lua_gettop(L);
	lua_pushcfunction(L, trap_call_cb);
	lua_setfield(L, metatable, "__call");
	lua_setmetatable(L, trap_index);

	return 0;
}

static int
log_info(lua_State *L)
{
	int n = lua_gettop(L);
	int i;
	lua_getfield(L, LUA_REGISTRYINDEX, "handlerptr");
	proc_handler *h = (proc_handler *)lua_touserdata(L, lua_gettop(L));
	lua_pop(L, 1);
	for (i = 1; i <= n; i++) {
		log_handler(LOG_INFO, h->name, "%s", lua_tostring(L, i));
	}
	return 0;
}

static int
dump_lua_trap(lua_State *L)
{
	lua_getglobal(L, "trap");
	lua_getfield(L, LUA_REGISTRYINDEX, "objptr");
	int point = lua_gettop(L);
	proc_trap *trap = (proc_trap *)lua_touserdata(L, point);
	lua_getfield(L, LUA_REGISTRYINDEX, "handlerptr");
	proc_handler *h = (proc_handler *)lua_touserdata(L, lua_gettop(L));
	lua_pop(L, 1);
	log_handler(LOG_ERR, h->name, "%s", display_trap(trap));
	return 0;
}

static int
action_done(lua_State *L)
{
	lua_pushstring(L, "DONE");
	return lua_error(L);
}

static int
action_ignore(lua_State *L)
{
	lua_pushstring(L, "IGNORED");
	return lua_error(L);
}

static int
action_reject(lua_State *L)
{
	lua_pushstring(L, "REJECTED");
	return lua_error(L);
}

static int
include(lua_State *L)
{
	int res, n = lua_gettop(L);
	if (n != 1)
		return luaL_error(L, "No module supplied to include");
	if (!lua_isstring(L, n))
		return luaL_error(L, "Can't include module - supplied argument not a string");
	const char *module = lua_tostring(L, n);
	lua_pop(L, 1);
	char *script = get_module(module);
	if (script == NULL) {
		return luaL_error(L, "Can't find module '%s'", module);
	}
	if ((res = luaL_loadstring(L, script))) {
		switch (res) {
			case LUA_ERRSYNTAX:
				return luaL_error(L, "Syntax error loading module %s: %s", module, script);
				break;
			case LUA_ERRMEM:
				return luaL_error(L, "Out of memory when loading module %s", module);
				break;
			default:
				return luaL_error(L, "Unknown error when loading module %s", module);
				break;
		}
	}
	int modidx = lua_gettop(L);
	lua_getfield(L, LUA_REGISTRYINDEX, "theenv");
	lua_setfenv(L, modidx);
	if ((res = lua_pcall(L, 0, LUA_MULTRET, 0))) {
		switch (res) {
			case LUA_ERRRUN:
				return luaL_error(L, "Runtime error running module %s: %s", module, script);
				break;
			case LUA_ERRMEM:
				return luaL_error(L, "Out of memory when running module %s", module);
				break;
			case LUA_ERRERR:
				return luaL_error(L, "Error in error handler when running module %s", module);
				break;
			default:
				return luaL_error(L, "Unknown error when running module %s", module);
				break;
		}
	}
	free(script);
	return lua_gettop(L);
}

static void
init_lua_env(lua_State *L)
{
	char *builtins[] = {
		/* base? */
		"ipairs",
		"next",
		"pairs",
		"select",
		"tonumber",
		"tostring",
		"type",
		"unpack",
		/* and specific libs */
		"string",
		"table",
		"math",
		NULL,
	};
	int i = 0;
	lua_newtable(L);
	int env = lua_gettop(L);
	for (i = 0; builtins[i] != NULL; i++) {
		lua_getglobal(L, builtins[i]);
		lua_setfield(L, env, builtins[i]);
	}
	lua_getglobal(L, "constant");
	lua_newtable(L);
	int state = lua_gettop(L);
		lua_pushnumber(L, STATE_UNCHANGED);
		lua_setfield(L, state, "UNCHANGED");
		lua_pushnumber(L, STATE_OK);
		lua_setfield(L, state, "OK");
		lua_pushnumber(L, STATE_WARNING);
		lua_setfield(L, state, "WARNING");
		lua_pushnumber(L, STATE_CRITICAL);
		lua_setfield(L, state, "CRITICAL");
		lua_pushnumber(L, STATE_UNKNOWN);
		lua_setfield(L, state, "UNKNOWN");
	lua_call(L, 1, 1);
	lua_setfield(L, env, "STATE");

	lua_getglobal(L, "constant");
	lua_newtable(L);
	int notify = lua_gettop(L);
		lua_pushnumber(L, NOTIFY_NEW_STATE);
		lua_setfield(L, notify, "NEW_STATE");
		lua_pushnumber(L, NOTIFY_ALWAYS);
		lua_setfield(L, notify, "ALWAYS");
	lua_call(L, 1, 1);
	lua_setfield(L, env, "NOTIFY");

	lua_pushcfunction(L, log_info);
	lua_setfield(L, env, "log");

	lua_getglobal(L, "constant");
	lua_newtable(L);
	int action = lua_gettop(L);
		lua_pushcfunction(L, action_done);
		lua_setfield(L, action, "done");
		lua_pushcfunction(L, action_ignore);
		lua_setfield(L, action, "ignore");
		lua_pushcfunction(L, action_reject);
		lua_setfield(L, action, "reject");
	lua_call(L, 1, 1);
	lua_setfield(L, env, "action");

	lua_pushcfunction(L, include);
	lua_setfield(L, env, "include");

	lua_getglobal(L, "constant");
	lua_newtable(L);
	int dump = lua_gettop(L);
		lua_pushcfunction(L, dump_lua_trap);
		lua_setfield(L, dump, "trap");
	lua_call(L, 1, 1);
	lua_setfield(L, env, "dump");
}

static char *
serialize(lua_State *L, int id) {
	proc_handler *h;
	switch(lua_type(L, id)) {
		case LUA_TNONE:
			lua_getfield(L, LUA_REGISTRYINDEX, "handlerptr");
			h = (proc_handler *)lua_touserdata(L, lua_gettop(L));
			lua_pop(L, 1);
			log_handler(LOG_ERR, h->name, "Invalid context index, returning empty table");
			return strdup("{}");
			break;
		case LUA_TNIL:
			return strdup("nil");
			break;
		case LUA_TNUMBER:
			{
				double n = lua_tonumber(L, id);
				char *res;
				if (asprintf(&res, "%g", n) < 0)
					return NULL;
				return res;
			}
			break;
		case LUA_TBOOLEAN:
			{
				int n = lua_toboolean(L, id);
				return strdup(n?"true":"false");
			}
			break;
		case LUA_TSTRING:
			{
				char *ret;
				const char *str = lua_tostring(L, id);
				ret = escape(str);
				return ret;
			}
			break;
		case LUA_TTABLE:
			{
				char *prev, *result = strdup("{");
				lua_pushnil(L);
				while (lua_next(L, id) != 0) {
					prev = result;
					char *val = serialize(L, lua_gettop(L));
					char *key = serialize(L, lua_gettop(L)-1);
					lua_pop(L, 1);
					if (asprintf(&result, "%s[%s]=%s,", prev, key, val) < 0)
						return NULL;
					free(prev);
					free(key);
					free(val);
				}
				if (strlen(result) == 1) {
					free(result);
					result = strdup("{}");
				} else {
					result[strlen(result)-1] = '}';
				}
				return result;
			}
			break;
		case LUA_TFUNCTION:
		case LUA_TUSERDATA:
		case LUA_TTHREAD:
		case LUA_TLIGHTUSERDATA:
			/* possible, but makes no sense */
			return strdup("{}");
			break;
		default:
			lua_getfield(L, LUA_REGISTRYINDEX, "handlerptr");
			h = (proc_handler *)lua_touserdata(L, lua_gettop(L));
			lua_pop(L, 1);
			log_handler(LOG_ERR, h->name, "Unknown lua type to serialize, returning empty table");
			return strdup("{}");
			break;
	}
}

static int
handle_context(lua_State *L, unsigned int handler_id, int context)
{
	char *ctx = serialize(L, context);
	int res = save_context(ctx, handler_id);
	free(ctx);
	return res;
}

int
run_handler(proc_trap *trap, proc_handler *handler, proc_result **result)
{
	log_handler(LOG_INFO, handler->name, "Running handler");
	char *ctx = NULL;
	int res, top;
	lua_State *L = luaL_newstate();

	if (handler->context && strlen(handler->context)) {
		if (asprintf(&ctx, "return %s", handler->context) < 0)
			return RESULT_IGNORE_RESULT;

		if ((res = luaL_dostring(L, ctx))) {
			switch (res) {
				case LUA_ERRSYNTAX:
					log_handler(LOG_ERR, handler->name, "Syntax error loading context: %s", ctx);
					break;
				case LUA_ERRMEM:
					log_handler(LOG_ERR, handler->name, "Out of memory when loading context");
					break;
				default:
					log_handler(LOG_ERR, handler->name, "Unknown error when loading context: %s", ctx);
					break;
			}
			/* ...but we don't return - corrupt context gets deleted and a new context is put there */
			lua_pop(L, 1);
			lua_newtable(L);
		}
	}
	/* no/invalid context? ignore */
	if (lua_gettop(L) != 1 || lua_type(L, 1) != LUA_TTABLE) {
		lua_pop(L, lua_gettop(L));
		lua_newtable(L);
	}
	lua_pushlightuserdata(L, trap);
	lua_setfield(L, LUA_REGISTRYINDEX, "objptr");
	lua_pushlightuserdata(L, handler);
	lua_setfield(L, LUA_REGISTRYINDEX, "handlerptr");
	lua_setfield(L, LUA_REGISTRYINDEX, "thecontext");

	if ((res = luaL_loadstring(L, handler->script))) {
		switch (res) {
			case LUA_ERRSYNTAX:
				log_handler(LOG_ERR, handler->name, "Syntax error loading handlers: %s", handler->script);
				break;
			case LUA_ERRMEM:
				log_handler(LOG_ERR, handler->name, "Out of memory when loading handler");
				break;
			default:
				log_handler(LOG_ERR, handler->name, "Unknown error when loading handler");
				break;
		}
		lua_close(L);
		return RESULT_IGNORE_RESULT;
	}
	int scriptidx = lua_gettop(L);

	luaL_openlibs(L);

	/* this file makes string.lines and string.trim available to
	 * handlers. it's not documented anywhere, but that's what happens.
	 * jay :/
	 */

	/* this file should always be there, so abort if it isn't - we might even want to exit() */
	char *extrafiles[] = {
		"share/processor/extra.lua",
		PREFIX "/share/processor/extra.lua",
		NULL
	};
	int i;
	struct stat extra;
	for (i = 0; extrafiles[i]; i++) {
		if (!stat(extrafiles[i], &extra))
			break;
	}

	if (!extrafiles[i]) {
		log_handler(LOG_ERR, handler->name, "Couldn't find file extra.lua");
		return RESULT_IGNORE_RESULT | RESULT_STOP_HANDLING;
	}
	res = luaL_loadfile(L, extrafiles[i]);
	if (res) {
		switch (res) {
			case LUA_ERRSYNTAX:
				log_handler(LOG_ERR, handler->name, "Syntax error loading extra.lua");
				break;
			case LUA_ERRMEM:
				log_handler(LOG_ERR, handler->name, "Out of memory when loading extra.lua");
				break;
			case LUA_ERRFILE:
				log_handler(LOG_ERR, handler->name, "Couldn't find file extra.lua");
				break;
			default:
				log_handler(LOG_ERR, handler->name, "Unknown error when loading extra.lua");
				break;
		}
		lua_close(L);
		return RESULT_IGNORE_RESULT | RESULT_STOP_HANDLING;
	}
	if ((res = lua_pcall(L, 0, LUA_MULTRET, 0))) {
		switch (res) {
			case LUA_ERRRUN:
				log_handler(LOG_ERR, handler->name, "Error running extra.lua");
				break;
			case LUA_ERRMEM:
				log_handler(LOG_ERR, handler->name, "Out of memory when running extra.lua");
				break;
			case LUA_ERRERR:
				log_handler(LOG_ERR, handler->name, "An error occured in the error handler for extra.lua");
				break;
			default:
				log_handler(LOG_ERR, handler->name, "Unknown error when running extra.lua");
				break;
		}
		lua_close(L);
		return RESULT_IGNORE_RESULT | RESULT_STOP_HANDLING;
	}

	init_lua_env(L);
	top = lua_gettop(L);

	lua_getfield(L, LUA_REGISTRYINDEX, "thecontext");
	lua_setfield(L, top, "context");

	lua_getglobal(L, "constant");
	trap_as_table(L, trap);
	lua_call(L, 1, 1);
	lua_setfield(L, top, "trap");

	lua_newtable(L);
	int resultidx = lua_gettop(L);
	lua_pushstring(L, trap->host);
	lua_setfield(L, resultidx, "host");
	lua_pushstring(L, "SNMP Traps and Notifications");
	lua_setfield(L, resultidx, "service");
	lua_pushstring(L, "default");
	lua_setfield(L, resultidx, "tag");
	lua_pushstring(L, "");
	lua_setfield(L, resultidx, "message");
	lua_pushnumber(L, STATE_UNCHANGED);
	lua_setfield(L, resultidx, "state");
	lua_pushnumber(L, NOTIFY_NEW_STATE);
	lua_setfield(L, resultidx, "notify");
	lua_setfield(L, top, "result");

	lua_setfield(L, LUA_REGISTRYINDEX, "theenv");
	lua_getfield(L, LUA_REGISTRYINDEX, "theenv");

	lua_setfenv(L, scriptidx);

	if ((res = lua_pcall(L, 0, LUA_MULTRET, 0))) {
		switch (res) {
			case LUA_ERRMEM:
				log_handler(LOG_ERR, handler->name, "Memory error running handler");
				return RESULT_STOP_HANDLING | RESULT_IGNORE_RESULT;
				break;
			case LUA_ERRERR:
				log_handler(LOG_ERR, handler->name, "Error running error handler");
				return RESULT_STOP_HANDLING | RESULT_IGNORE_RESULT;
				break;
			case LUA_ERRRUN:
				{
					const char *msg = lua_tostring(L, lua_gettop(L));
					if (!msg) {
						log_handler(LOG_ERR, handler->name, "An error occured during execution, but no error message was pushed");
						return RESULT_STOP_HANDLING | RESULT_IGNORE_RESULT;
					}

					if (!strcmp(msg, "DONE")) {
						log_handler(LOG_DEBUG, handler->name, "Handler called done on trap");
						/* the no-error error - that is, this does a return from the handler
						 * this is not IGNORED, because this processes the handler result */
						lua_pop(L, 1);
						/* darn, out of stack :( */
						lua_getfield(L, LUA_REGISTRYINDEX, "theenv");
					}
					else if (!strcmp(msg, "REJECTED")) {
						/* the stop processing traps error - that is, break from trap processing */
						log_handler(LOG_DEBUG, handler->name, "Handler rejected trap");
						return RESULT_STOP_HANDLING | RESULT_IGNORE_RESULT;
					}
					else if (!strcmp(msg, "IGNORED")) {
						/* the stop-with-this-handler-but-go-on-with-the-next-one - that is, continue in trap processing */
						log_handler(LOG_DEBUG, handler->name, "Handler ignored trap");
						return RESULT_IGNORE_RESULT;
					}
					else {
						log_handler(LOG_ERR, handler->name, "An unknown error was thrown during execution: %s", msg);
						return RESULT_STOP_HANDLING | RESULT_IGNORE_RESULT;
					}
				}
				break;
			default:
				log_handler(LOG_ERR, handler->name, "An unknown error occured during execution");
				return RESULT_STOP_HANDLING | RESULT_IGNORE_RESULT;
				break;
		}
	} else {
		lua_getfenv(L, lua_gettop(L));
	}

	top = lua_gettop(L);
	lua_getfield(L, top, "result");
	resultidx = lua_gettop(L);

	if (result) {
		if (lua_type(L, resultidx) != LUA_TTABLE) {
			log_handler(LOG_ERR, handler->name, "Invalid result - not a table");
			return RESULT_IGNORE_RESULT;
		}
		const char *str;
		*result = calloc(1, sizeof(proc_result));
		lua_getfield(L, resultidx, "host");
		str = lua_tostring(L, lua_gettop(L));
		if (!str)
			str = "";
		(*result)->host = strdup(str);
		lua_getfield(L, resultidx, "service");
		str = lua_tostring(L, lua_gettop(L));
		if (!str)
			str = "";
		(*result)->service = strdup(str);
		lua_getfield(L, resultidx, "tag");
		str = lua_tostring(L, lua_gettop(L));
		if (!str)
			str = "";
		(*result)->tag = strdup(str);
		lua_getfield(L, resultidx, "state");
		(*result)->state = (int)lua_tonumber(L, lua_gettop(L));
		lua_getfield(L, resultidx, "message");
		str = lua_tostring(L, lua_gettop(L));
		if (!str)
			str = "";
		(*result)->message = strdup(str);
		lua_getfield(L, resultidx, "notify");
		(*result)->notify = (int)lua_tonumber(L, lua_gettop(L));
		log_handler(LOG_DEBUG, handler->name, "Result: %s;%s;%s = %d (%s)", (*result)->host ? (*result)->host : "(NULL)", (*result)->service ? (*result)->service : "(NULL)", (*result)->tag ? (*result)->tag : "(NULL)", (*result)->state, (*result)->message ? (*result)->message : "(NULL)");

		lua_getfield(L, LUA_REGISTRYINDEX, "objptr");
		(*result)->trap = (proc_trap *)lua_topointer(L, lua_gettop(L));
		(*result)->handler = handler;
	}

	lua_getfield(L, top, "context");
	handle_context(L, trap->id, lua_gettop(L));
	if (ctx) {
		free(ctx);
		ctx = NULL;
	}
	lua_close(L);
	return RESULT_CONTINUE;
}
